# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sigstore based signature, signers and verifiers."""

import pathlib
import sys
from typing import Optional, cast

from google.protobuf import json_format
from sigstore import dsse as sigstore_dsse
from sigstore import models as sigstore_models
from sigstore import oidc as sigstore_oidc
from sigstore import sign as sigstore_signer
from sigstore import verify as sigstore_verifier
from typing_extensions import override

from model_signing._signing import signing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

_DEFAULT_CLIENT_ID = "sigstore"
_DEFAULT_CLIENT_SECRET = ""


class Signature(signing.Signature):
    """Sigstore signature support, wrapping around `sigstore_models.Bundle`."""

    def __init__(self, bundle: sigstore_models.Bundle):
        """Builds an instance of this signature.

        Args:
            bundle: the sigstore bundle (in `bundle_pb.Bundle` format).
        """
        self.bundle = bundle

    @override
    def write(self, path: pathlib.Path) -> None:
        path.write_text(self.bundle.to_json(), encoding="utf-8")

    @classmethod
    @override
    def read(cls, path: pathlib.Path) -> Self:
        content = path.read_text(encoding="utf-8")
        return cls(sigstore_models.Bundle.from_json(content))


class Signer(signing.Signer):
    """Signing using Sigstore."""

    def __init__(
        self,
        *,
        oidc_issuer: Optional[str] = None,
        use_ambient_credentials: bool = True,
        use_staging: bool = False,
        identity_token: Optional[str] = None,
        force_oob: bool = False,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        trust_config: Optional[pathlib.Path] = None,
    ):
        """Initializes Sigstore signers.

        Needs to set-up a signing context to use the public goods instance and
        machinery for getting an identity token to use in signing.

        Args:
            oidc_issuer: An optional OpenID Connect issuer to use instead of the
              default production one. Only relevant if `use_staging = False`.
              Default is empty, relying on the Sigstore configuration.
            use_ambient_credentials: Use ambient credentials (also known as
              Workload Identity). Default is True. If ambient credentials cannot
              be used (not available, or option disabled), a flow to get signer
              identity via OIDC will start.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
            force_oob: If True, forces an out-of-band (OOB) OAuth flow. If set,
              the OAuth authentication will not attempt to open the default web
              browser. Instead, it will display a URL and code for manual
              authentication. Default is False, which means the browser will be
              opened automatically if possible.
            identity_token: An explicit identity token to use when signing,
              taking precedence over any ambient credential or OAuth workflow.
             client_id: An optional client ID to use when performing OIDC-based
              authentication. This is typically used to identify the
              application making the request to the OIDC provider. If not
              provided, the default client ID configured by Sigstore will be
              used.
            client_secret: An optional client secret to use along with the
              client ID when authenticating with the OIDC provider. This is
              required for confidential clients that need to prove their
              identity to the OIDC provider. If not provided, it is assumed
              that the client is public or the provider does not require a
              secret.
            trust_config: A path to a custom trust configuration. When
              provided, the signature verification process will rely on the
              supplied PKI and trust configurations, instead of the default
              Sigstore setup. If not specified, the default Sigstore
              configuration is used.
        """
        if use_staging:
            trust_config = sigstore_models.ClientTrustConfig.staging()
        elif trust_config:
            trust_config = sigstore_models.ClientTrustConfig.from_json(
                trust_config.read_text()
            )
        else:
            trust_config = sigstore_models.ClientTrustConfig.production()

        if not oidc_issuer:
            oidc_issuer = trust_config.signing_config.get_oidc_url()

        self._issuer = sigstore_oidc.Issuer(oidc_issuer)
        self._signing_context = (
            sigstore_signer.SigningContext.from_trust_config(trust_config)
        )
        self._use_ambient_credentials = use_ambient_credentials
        self._identity_token = identity_token
        self._force_oob = force_oob
        self._client_id = client_id or _DEFAULT_CLIENT_ID
        self._client_secret = client_secret or _DEFAULT_CLIENT_SECRET

    def _get_identity_token(self) -> sigstore_oidc.IdentityToken:
        """Obtains an identity token to use in signing.

        The precedence matches that of sigstore-python:
        1) Explicitly supplied identity token
        2) Ambient credential detected in the environment, if enabled
        3) Interactive OAuth flow
        """
        if self._identity_token:
            return sigstore_oidc.IdentityToken(
                self._identity_token, self._client_id
            )
        if self._use_ambient_credentials:
            token = sigstore_oidc.detect_credential(self._client_id)
            if token:
                return sigstore_oidc.IdentityToken(token, self._client_id)

        return self._issuer.identity_token(
            force_oob=self._force_oob,
            client_id=self._client_id,
            client_secret=self._client_secret,
        )

    @override
    def sign(self, payload: signing.Payload) -> Signature:
        # We need to convert from in-toto statement to Sigstore's DSSE
        # version. They both contain the same contents, but there is no way
        # to coerce one type to the other.
        # See also: https://github.com/sigstore/sigstore-python/issues/1076
        statement = sigstore_dsse.Statement(
            json_format.MessageToJson(payload.statement.pb).encode("utf-8")
        )

        token = self._get_identity_token()
        with self._signing_context.signer(token) as signer:
            bundle = signer.sign_dsse(statement)

        return Signature(bundle)


class Verifier(signing.Verifier):
    """Signature verification using Sigstore."""

    def __init__(
        self,
        *,
        identity: str,
        oidc_issuer: str,
        use_staging: bool = False,
        trust_config: Optional[pathlib.Path] = None,
    ):
        """Initializes Sigstore verifiers.

        When verifying a signature, we also check an identity policy: the
        certificate must belong to a given "identity", and must be issued by a
        given OpenID Connect issuer.

        Args:
            identity: The expected identity that has signed the model.
            oidc_issuer: The expected OpenID Connect issuer that provided the
              certificate used for the signature.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
            trust_config: A path to a custom trust configuration. When provided,
              the signature verification process will rely on the supplied
              PKI and trust configurations, instead of the default Sigstore
              setup. If not specified, the default Sigstore configuration
              is used.
        """
        if trust_config:
            trust_config = sigstore_models.ClientTrustConfig.from_json(
                trust_config.read_text()
            )
        elif use_staging:
            trust_config = sigstore_models.ClientTrustConfig.staging()
        else:
            trust_config = sigstore_models.ClientTrustConfig.production()

        self._verifier = sigstore_verifier.Verifier(
            trusted_root=trust_config.trusted_root
        )

        self._policy = sigstore_verifier.policy.Identity(
            identity=identity, issuer=oidc_issuer
        )

    @override
    def _verify_signed_content(
        self, signature: signing.Signature
    ) -> tuple[str, bytes]:
        # We are guaranteed to only use the local signature type
        signature = cast(Signature, signature)
        bundle = signature.bundle
        return self._verifier.verify_dsse(bundle=bundle, policy=self._policy)



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Digests for memory objects.

These can only compute hashes of objects residing in memory, after they get
converted to bytes.

Example usage:
```python
>>> hasher = SHA256()
>>> hasher.update(b"abcd")
>>> digest = hasher.compute()
>>> digest.digest_hex
'88d4266fd4e6338d13b845fcf289579d209c897823b9217da3e161936f031589'
```

Or, passing the data directly in the constructor:
```python
>>> hasher = SHA256(b"abcd")
>>> digest = hasher.compute()
>>> digest.digest_hex
'88d4266fd4e6338d13b845fcf289579d209c897823b9217da3e161936f031589'
```
"""

import hashlib

import blake3
from typing_extensions import override

from model_signing._hashing import hashing


class SHA256(hashing.StreamingHashEngine):
    """A wrapper around `hashlib.sha256`."""

    def __init__(self, initial_data: bytes = b""):
        """Initializes an instance of a SHA256 hash engine.

        Args:
            initial_data: Optional initial data to hash.
        """
        self._hasher = hashlib.sha256(initial_data)

    @override
    def update(self, data: bytes) -> None:
        self._hasher.update(data)

    @override
    def reset(self, data: bytes = b"") -> None:
        self._hasher = hashlib.sha256(data)

    @override
    def compute(self) -> hashing.Digest:
        return hashing.Digest(self.digest_name, self._hasher.digest())

    @property
    @override
    def digest_name(self) -> str:
        return "sha256"

    @property
    @override
    def digest_size(self) -> int:
        return self._hasher.digest_size


class BLAKE2(hashing.StreamingHashEngine):
    """A wrapper around `hashlib.blake2b`."""

    def __init__(self, initial_data: bytes = b""):
        """Initializes an instance of a BLAKE2 hash engine.

        Args:
            initial_data: Optional initial data to hash.
        """
        self._hasher = hashlib.blake2b(initial_data)

    @override
    def update(self, data: bytes) -> None:
        self._hasher.update(data)

    @override
    def reset(self, data: bytes = b"") -> None:
        self._hasher = hashlib.blake2b(data)

    @override
    def compute(self) -> hashing.Digest:
        return hashing.Digest(self.digest_name, self._hasher.digest())

    @property
    @override
    def digest_name(self) -> str:
        return "blake2b"

    @property
    @override
    def digest_size(self) -> int:
        return self._hasher.digest_size


class BLAKE3(hashing.StreamingHashEngine):
    """A wrapper around `blake3.blake3`."""

    def __init__(self, initial_data: bytes = b""):
        """Initializes an instance of a BLAKE3 hash engine.

        Args:
            initial_data: Optional initial data to hash.
        """
        self._hasher = blake3.blake3(initial_data)

    @override
    def update(self, data: bytes) -> None:
        self._hasher.update(data)

    @override
    def reset(self, data: bytes = b"") -> None:
        self._hasher = blake3.blake3(data)

    @override
    def compute(self) -> hashing.Digest:
        return hashing.Digest(self.digest_name, self._hasher.digest())

    @property
    @override
    def digest_name(self) -> str:
        return "blake3"

    @property
    @override
    def digest_size(self) -> int:
        return self._hasher.digest_size



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Machinery for computing digests for a single file.

Example usage for `SimpleFileHasher`:
```python
>>> with open("/tmp/file", "w") as f:
...     f.write("abcd")
>>> hasher = SimpleFileHasher("/tmp/file", SHA256())
>>> digest = hasher.compute()
>>> digest.digest_hex
'88d4266fd4e6338d13b845fcf289579d209c897823b9217da3e161936f031589'
```

Example usage for `ShardedFileHasher`, reading only the second part of a file:
```python
>>> with open("/tmp/file", "w") as f:
...     f.write("0123abcd")
>>> hasher = ShardedFileHasher("/tmp/file", SHA256(), start=4, end=8)
>>> digest = hasher.compute()
>>> digest.digest_hex
'88d4266fd4e6338d13b845fcf289579d209c897823b9217da3e161936f031589'
```
"""

import pathlib
from typing import Optional

import blake3
from typing_extensions import override

from model_signing._hashing import hashing


class FileHasher(hashing.HashEngine):
    """Generic file hash engine.

    This class is intentionally empty (and abstract, via inheritance) to be used
    only as a type annotation (to signal that API expects a hasher capable of
    hashing files, instead of any `HashEngine` instance).
    """


class SimpleFileHasher(FileHasher):
    """Simple file hash engine that computes the digest iteratively.

    To compute the hash of a file, we read the file exactly once, including for
    very large files that don't fit in memory. Files are read in chunks and each
    chunk is passed to the `update` method of an inner
    `hashing.StreamingHashEngine`, instance. This ensures that the file digest
    will not change even if the chunk size changes. As such, we can dynamically
    determine an optimal value for the chunk argument.
    """

    def __init__(
        self,
        file: pathlib.Path,
        content_hasher: hashing.StreamingHashEngine,
        *,
        chunk_size: int = 1_048_576,
        digest_name_override: Optional[str] = None,
    ):
        """Initializes an instance to hash a file with a specific `HashEngine`.

        Args:
            file: The file to hash. Use `set_file` to reset it.
            content_hasher: A `hashing.StreamingHashEngine` instance used to
              compute the digest of the file.
            chunk_size: The amount of file to read at once. Default is 1MB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            digest_name_override: Optional string to allow overriding the
              `digest_name` property to support shorter, standardized names.
        """
        if chunk_size < 0:
            raise ValueError(
                f"Chunk size must be non-negative, got {chunk_size}."
            )

        self._file = file
        self._content_hasher = content_hasher
        self._chunk_size = chunk_size
        self._digest_name_override = digest_name_override

    def set_file(self, file: pathlib.Path) -> None:
        """Redefines the file to be hashed in `compute`.

        Args:
            file: The new file to be hashed.
        """
        self._file = file

    @property
    @override
    def digest_name(self) -> str:
        if self._digest_name_override is not None:
            return self._digest_name_override
        # Since there is no difference between hashing the file with this engine
        # or reading the file in memory and then using the content hasher
        # directly, we must have the same digest_name.
        return self._content_hasher.digest_name

    @override
    def compute(self) -> hashing.Digest:
        self._content_hasher.reset()

        if self._chunk_size == 0:
            with open(self._file, "rb") as f:
                self._content_hasher.update(f.read())
        else:
            with open(self._file, "rb") as f:
                while True:
                    data = f.read(self._chunk_size)
                    if not data:
                        break
                    self._content_hasher.update(data)

        digest = self._content_hasher.compute()
        return hashing.Digest(self.digest_name, digest.digest_value)

    @property
    @override
    def digest_size(self) -> int:
        return self._content_hasher.digest_size


class Blake3FileHasher(FileHasher):
    """Simple file hash engine that uses BLAKE3 in parallel.

    This hash engine uses the fastest BLAKE3 settings, by using memory mapping
    and multiple workers. This will greatly increase speed on SSDs, but may
    not perform well on HDDs. For HDDs, you can set max_threads to 1.
    """

    def __init__(
        self,
        file: pathlib.Path,
        *,
        max_threads: int = blake3.blake3.AUTO,
        digest_name_override: Optional[str] = None,
    ):
        """Initializes an instance to hash a file.

        Args:
            file: The file to hash. Use `set_file` to reset it.
            max_threads: how many BLAKE3 workers to use. Defaults to number of
              logical cores.
            digest_name_override: Optional string to allow overriding the
              `digest_name` property to support shorter, standardized names.
        """
        self._file = file
        self._digest_name_override = digest_name_override
        self._blake3 = blake3.blake3(max_threads=max_threads)

    def set_file(self, file: pathlib.Path) -> None:
        """Redefines the file to be hashed in `compute`.

        Args:
            file: The new file to be hashed.
        """
        self._file = file

    @property
    @override
    def digest_name(self) -> str:
        if self._digest_name_override is not None:
            return self._digest_name_override
        return "blake3"

    @override
    def compute(self) -> hashing.Digest:
        self._blake3.reset()
        self._blake3.update_mmap(self._file)
        return hashing.Digest(self.digest_name, self._blake3.digest())

    @property
    @override
    def digest_size(self) -> int:
        return 32


class ShardedFileHasher(SimpleFileHasher):
    """File hash engine that hashes a portion (shard) of the file.

    By invoking this engine in parallel across disjoint shards, we can speed up
    hashing a single file. However, the hash output depends on the shard size.

    It is the responsibility of the user to compose the digests of each shard
    into a single digest for the entire file.
    """

    def __init__(
        self,
        file: pathlib.Path,
        content_hasher: hashing.StreamingHashEngine,
        *,
        start: int,
        end: int,
        chunk_size: int = 1_048_576,
        shard_size: int = 1_000_000_000,
        digest_name_override: Optional[str] = None,
    ):
        """Initializes an instance to hash a file with a specific `HashEngine`.

        Args:
            file: The file to hash. Use `set_file` to reset it.
            content_hasher: A `hashing.HashEngine` instance used to compute the
              digest of the file shard.
            start: The file offset to start reading from. Must be valid. Reset
              with `set_shard`.
            end: The file offset to stop reading at. Must be stricly greater
              than start. The entire shard length must be less than the
              configured `shard_size`. Reset with `set_shard`.
            chunk_size: The amount of file to read at once. Default is 1MB. A
              special value of 0 signals to attempt to read everything in a
              single call.
            shard_size: The size of a file shard. Default is 1 GB.
            digest_name_override: Optional string to allow overriding the
              `digest_name` property to support shorter, standardized names.
        """
        super().__init__(
            file=file,
            content_hasher=content_hasher,
            chunk_size=chunk_size,
            digest_name_override=digest_name_override,
        )

        if shard_size <= 0:
            raise ValueError(
                f"Shard size must be strictly positive, got {shard_size}."
            )
        self.shard_size = shard_size

        self.set_shard(start=start, end=end)

    def set_shard(self, *, start: int, end: int) -> None:
        """Redefines the file shard to be hashed in `compute`.

        Args:
            start: The file offset to start reading from. Must be valid.
            end: The file offset to stop reading at. Must be stricly greater
              than start. The entire shard length must be less than the
              configured `shard_size`.
        """
        if start < 0:
            raise ValueError(
                f"File start offset must be non-negative, got {start}."
            )
        if end <= start:
            raise ValueError(
                "File end offset must be stricly higher that file start offset,"
                f" got {start=}, {end=}."
            )
        read_length = end - start
        if read_length > self.shard_size:
            raise ValueError(
                f"Must not read more than shard_size={self.shard_size}, got"
                f" {read_length}."
            )

        self._start = start
        self._end = end

    @override
    def compute(self) -> hashing.Digest:
        self._content_hasher.reset()

        with open(self._file, "rb") as f:
            f.seek(self._start)
            to_read = self._end - self._start
            if self._chunk_size == 0 or self._chunk_size >= to_read:
                data = f.read(to_read)
                self._content_hasher.update(data)
            else:
                while to_read >= 0:
                    data = f.read(min(self._chunk_size, to_read))
                    if not data:
                        break
                    to_read -= len(data)
                    self._content_hasher.update(data)

        digest = self._content_hasher.compute()
        return hashing.Digest(self.digest_name, digest.digest_value)

    @property
    @override
    def digest_name(self) -> str:
        if self._digest_name_override is not None:
            return self._digest_name_override
        return f"{self._content_hasher.digest_name}-sharded-{self.shard_size}"



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""High level API for the verification interface of `model_signing` library.

This module supports configuring the verification method used to verify a model,
before performing the verification.

```python
model_signing.verifying.Config().use_sigstore_verifier(
    identity=identity, oidc_issuer=oidc_provider
).verify("finbert", "finbert.sig")
```

The same verification configuration can be used to verify multiple models:

```python
verifying_config = model_signing.signing.Config().use_elliptic_key_verifier(
    public_key="key.pub"
)

for model in all_models:
    verifying_config.verify(model, f"{model}_sharded.sig")
```

The API defined here is stable and backwards compatible.
"""

from collections.abc import Iterable
import pathlib
import sys
from typing import Optional

from model_signing import hashing
from model_signing import manifest
from model_signing._signing import sign_certificate as certificate
from model_signing._signing import sign_ec_key as ec_key
from model_signing._signing import sign_sigstore as sigstore
from model_signing._signing import sign_sigstore_pb as sigstore_pb


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


class Config:
    """Configuration to use when verifying models against signatures.

    The verification configuration is needed to determine how to read and verify
    the signature. Given we support multiple signing format, the verification
    settings must match the signing ones.

    The configuration also supports configuring the hashing configuration from
    `model_signing.hashing`. This should also match the configuration used
    during signing. However, by default, we can attempt to guess it from the
    signature.
    """

    def __init__(self):
        """Initializes the default configuration for verification."""
        self._hashing_config = None
        self._verifier = None
        self._uses_sigstore = False
        self._ignore_unsigned_files = False

    def verify(
        self, model_path: hashing.PathLike, signature_path: hashing.PathLike
    ):
        """Verifies that a model conforms to a signature.

        Args:
            model_path: The path to the model to verify.

        Raises:
            ValueError: No verifier has been configured.
        """
        if self._verifier is None:
            raise ValueError("Attempting to verify with no configured verifier")

        if self._uses_sigstore:
            signature = sigstore.Signature.read(pathlib.Path(signature_path))
        else:
            signature = sigstore_pb.Signature.read(pathlib.Path(signature_path))

        expected_manifest = self._verifier.verify(signature)

        if self._hashing_config is None:
            self._guess_hashing_config(expected_manifest)
        if "ignore_paths" in expected_manifest.serialization_type:
            self._hashing_config.add_ignored_paths(
                model_path=model_path,
                paths=expected_manifest.serialization_type["ignore_paths"],
            )

        if self._ignore_unsigned_files:
            files_to_hash = [
                model_path / rd.identifier
                for rd in expected_manifest.resource_descriptors()
            ]
        else:
            files_to_hash = None

        actual_manifest = self._hashing_config.hash(
            model_path, files_to_hash=files_to_hash
        )

        if actual_manifest != expected_manifest:
            diff_message = self._get_manifest_diff(
                actual_manifest, expected_manifest
            )
            raise ValueError(f"Signature mismatch: {diff_message}")

    def _get_manifest_diff(self, actual, expected) -> list[str]:
        diffs = []

        actual_hashes = {
            rd.identifier: rd.digest for rd in actual.resource_descriptors()
        }
        expected_hashes = {
            rd.identifier: rd.digest for rd in expected.resource_descriptors()
        }

        extra_actual_files = set(actual_hashes.keys()) - set(
            expected_hashes.keys()
        )
        if extra_actual_files:
            diffs.append(
                f"Extra files found in model '{actual.model_name}': "
                f"{', '.join(sorted(extra_actual_files))}"
            )

        missing_actual_files = set(expected_hashes.keys()) - set(
            actual_hashes.keys()
        )
        if missing_actual_files:
            diffs.append(
                f"Missing files in model '{actual.model_name}': "
                f"{', '.join(sorted(missing_actual_files))}"
            )

        common_files = set(actual_hashes.keys()) & set(expected_hashes.keys())
        for identifier in sorted(common_files):
            if actual_hashes[identifier] != expected_hashes[identifier]:
                diffs.append(
                    f"Hash mismatch for '{identifier}': "
                    f"Expected '{expected_hashes[identifier]}', "
                    f"Actual '{actual_hashes[identifier]}'"
                )
        return diffs

    def set_hashing_config(self, hashing_config: hashing.Config) -> Self:
        """Sets the new configuration for hashing models.

        After calling this method, the automatic guessing of the hashing
        configuration used during signing is no longer possible from within one
        instance of this class.

        Args:
            hashing_config: The new hashing configuration.

        Returns:
            The new signing configuration.
        """
        self._hashing_config = hashing_config
        return self

    def set_ignore_unsigned_files(self, ignore_unsigned_files: bool) -> Self:
        """Sets whether files that were not signed are to be ignored.

        This method allows to ignore those files that are not part of the
        manifest and therefor were not originally signed.

        Args:
            ignore_unsigned_files: whether to ignore unsigned files
        """
        self._ignore_unsigned_files = ignore_unsigned_files
        return self

    def _guess_hashing_config(self, source_manifest: manifest.Manifest) -> None:
        """Attempts to guess the hashing config from a manifest."""
        args = source_manifest.serialization_type
        method = args["method"]
        # TODO: Once Python 3.9 support is deprecated revert to using `match`
        if method == "files":
            self._hashing_config = hashing.Config().use_file_serialization(
                hashing_algorithm=args["hash_type"],
                allow_symlinks=args["allow_symlinks"],
                ignore_paths=args.get("ignore_paths", frozenset()),
            )
        elif method == "shards":
            self._hashing_config = hashing.Config().use_shard_serialization(
                hashing_algorithm=args["hash_type"],
                shard_size=args["shard_size"],
                allow_symlinks=args["allow_symlinks"],
                ignore_paths=args.get("ignore_paths", frozenset()),
            )
        else:
            raise ValueError("Cannot guess the hashing configuration")

    def use_sigstore_verifier(
        self,
        *,
        identity: str,
        oidc_issuer: str,
        use_staging: bool = False,
        trust_config: Optional[pathlib.Path] = None,
    ) -> Self:
        """Configures the verification of signatures produced by Sigstore.

        The verifier in this configuration is changed to one that performs
        verification of Sigstore signatures (sigstore bundles signed by
        keyless signing via Sigstore).

        Args:
            identity: The expected identity that has signed the model.
            oidc_issuer: The expected OpenID Connect issuer that provided the
              certificate used for the signature.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
            trust_config: A path to a custom trust configuration. When provided,
              the signature verification process will rely on the supplied
              PKI and trust configurations, instead of the default Sigstore
              setup. If not specified, the default Sigstore configuration
              is used.

        Return:
            The new verification configuration.
        """
        self._uses_sigstore = True
        self._verifier = sigstore.Verifier(
            identity=identity,
            oidc_issuer=oidc_issuer,
            use_staging=use_staging,
            trust_config=trust_config,
        )
        return self

    def use_elliptic_key_verifier(
        self, *, public_key: hashing.PathLike
    ) -> Self:
        """Configures the verification of signatures generated by a private key.

        The verifier in this configuration is changed to one that performs
        verification of sgistore bundles signed by an elliptic curve private
        key. The public key used in the configuration must match the private key
        used during signing.

        Args:
            public_key: The path to the public key to verify with.

        Return:
            The new verification configuration.
        """
        self._uses_sigstore = False
        self._verifier = ec_key.Verifier(pathlib.Path(public_key))
        return self

    def use_certificate_verifier(
        self,
        *,
        certificate_chain: Iterable[hashing.PathLike] = frozenset(),
        log_fingerprints: bool = False,
    ) -> Self:
        """Configures the verification of signatures generated by a certificate.

        The verifier in this configuration is changed to one that performs
        verification of sgistore bundles signed by a signing certificate.

        Args:
            certificate_chain: Certificate chain to establish root of trust. If
              empty, the operating system's one is used.
            log_fingerprints: Log certificates' SHA256 fingerprints

        Return:
            The new verification configuration.
        """
        self._uses_sigstore = False
        self._verifier = certificate.Verifier(
            [pathlib.Path(c) for c in certificate_chain],
            log_fingerprints=log_fingerprints,
        )
        return self



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""High level API for the signing interface of `model_signing` library.

The module allows signing a model with a default configuration:

```python
model_signing.signing.sign("finbert", "finbert.sig")
```

The module allows customizing the signing configuration before signing:

```python
model_signing.signing.Config().use_elliptic_key_signer(private_key="key").sign(
    "finbert", "finbert.sig"
)
```

The same signing configuration can be used to sign multiple models:

```python
signing_config = model_signing.signing.Config().use_elliptic_key_signer(
    private_key="key"
)

for model in all_models:
    signing_config.sign(model, f"{model}_sharded.sig")
```

The API defined here is stable and backwards compatible.
"""

from collections.abc import Iterable
import pathlib
import sys
from typing import Optional

from model_signing import hashing
from model_signing._signing import sign_certificate as certificate
from model_signing._signing import sign_ec_key as ec_key
from model_signing._signing import sign_sigstore as sigstore
from model_signing._signing import signing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


def sign(model_path: hashing.PathLike, signature_path: hashing.PathLike):
    """Signs a model using the default configuration.

    In this default configuration we sign using Sigstore and the default hashing
    configuration from `model_signing.hashing`.

    The resulting signature is in the Sigstore bundle format.

    Args:
        model_path: the path to the model to sign.
        signature_path: the path of the resulting signature.
    """
    Config().sign(model_path, signature_path)


class Config:
    """Configuration to use when signing models.

    Currently, we support signing with Sigstore (both the public
    instance and staging instance), signing with private keys,
    signing with signing certificates, and signing with custom
    PKI configurations using the `--trust_config` option.
    This allows users to bring their own trust configuration
    to sign and verify models. Other signing modes may be
    added in the future.
    """

    def __init__(self):
        """Initializes the default configuration for signing."""
        self._hashing_config = hashing.Config()
        # lazy initialize default signer at signing to avoid network calls
        self._signer = None

    def sign(
        self, model_path: hashing.PathLike, signature_path: hashing.PathLike
    ):
        """Signs a model using the current configuration.

        Args:
            model_path: The path to the model to sign.
            signature_path: The path of the resulting signature.
        """
        if not self._signer:
            self._signer = self.use_sigstore_signer()
        manifest = self._hashing_config.hash(model_path)
        payload = signing.Payload(manifest)
        signature = self._signer.sign(payload)
        signature.write(pathlib.Path(signature_path))

    def set_hashing_config(self, hashing_config: hashing.Config) -> Self:
        """Sets the new configuration for hashing models.

        Args:
            hashing_config: The new hashing configuration.

        Returns:
            The new signing configuration.
        """
        self._hashing_config = hashing_config
        return self

    def use_sigstore_signer(
        self,
        *,
        oidc_issuer: Optional[str] = None,
        use_ambient_credentials: bool = False,
        use_staging: bool = False,
        force_oob: bool = False,
        identity_token: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
        trust_config: Optional[pathlib.Path] = None,
    ) -> Self:
        """Configures the signing to be performed with Sigstore.

        The signer in this configuration is changed to one that performs signing
        with Sigstore.

        Args:
            oidc_issuer: An optional OpenID Connect issuer to use instead of the
              default production one. Only relevant if `use_staging = False`.
              Default is empty, relying on the Sigstore configuration.
            use_ambient_credentials: Use ambient credentials (also known as
              Workload Identity). Default is False. If ambient credentials
              cannot be used (not available, or option disabled), a flow to get
              signer identity via OIDC will start.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
            force_oob: If True, forces an out-of-band (OOB) OAuth flow. If set,
              the OAuth authentication will not attempt to open the default web
              browser. Instead, it will display a URL and code for manual
              authentication. Default is False, which means the browser will be
              opened automatically if possible.
            identity_token: An explicit identity token to use when signing,
              taking precedence over any ambient credential or OAuth workflow.
            client_id: An optional client ID to use when performing OIDC-based
              authentication. This is typically used to identify the
              application making the request to the OIDC provider. If not
              provided, the default client ID configured by Sigstore will be
              used.
            client_secret: An optional client secret to use along with the
              client ID when authenticating with the OIDC provider. This is
              required for confidential clients that need to prove their
              identity to the OIDC provider. If not provided, it is assumed
              that the client is public or the provider does not require a
              secret.
            trust_config: A path to a custom trust configuration. When provided,
              the signature verification process will rely on the supplied
              PKI and trust configurations, instead of the default Sigstore
              setup. If not specified, the default Sigstore configuration
              is used.

        Return:
            The new signing configuration.
        """
        self._signer = sigstore.Signer(
            oidc_issuer=oidc_issuer,
            use_ambient_credentials=use_ambient_credentials,
            use_staging=use_staging,
            identity_token=identity_token,
            force_oob=force_oob,
            client_id=client_id,
            client_secret=client_secret,
            trust_config=trust_config,
        )
        return self

    def use_elliptic_key_signer(
        self, *, private_key: hashing.PathLike, password: Optional[str] = None
    ) -> Self:
        """Configures the signing to be performed using elliptic curve keys.

        The signer in this configuration is changed to one that performs signing
        using a private key based on elliptic curve cryptography.

        Args:
            private_key: The path to the private key to use for signing.
            password: An optional password for the key, if encrypted.

        Return:
            The new signing configuration.
        """
        self._signer = ec_key.Signer(pathlib.Path(private_key), password)
        return self

    def use_certificate_signer(
        self,
        *,
        private_key: hashing.PathLike,
        signing_certificate: hashing.PathLike,
        certificate_chain: Iterable[hashing.PathLike],
    ) -> Self:
        """Configures the signing to be performed using signing certificates.

        The signer in this configuration is changed to one that performs signing
        using cryptographic signing certificates.

        Args:
            private_key: The path to the private key to use for signing.
            signing_certificate: The path to the signing certificate.
            certificate_chain: Optional paths to other certificates to establish
              a chain of trust.

        Return:
            The new signing configuration.
        """
        self._signer = certificate.Signer(
            pathlib.Path(private_key),
            pathlib.Path(signing_certificate),
            [pathlib.Path(c) for c in certificate_chain],
        )
        return self

    def use_pkcs11_signer(
        self, *, pkcs11_uri: str, module_paths: Iterable[str] = frozenset()
    ) -> Self:
        """Configures the signing to be performed using PKCS #11.

        The signer in this configuration is changed to one that performs signing
        using a private key based on elliptic curve cryptography.

        Args:
            pkcs11_uri: The PKCS11 URI.
            module_paths: Optional list of paths of PKCS #11 modules.

        Return:
            The new signing configuration.
        """
        try:
            from model_signing._signing import sign_pkcs11 as pkcs11
        except ImportError as e:
            raise RuntimeError(
                "PKCS #11 functionality requires the 'pkcs11' extra. "
                "Install with 'pip install model-signing[pkcs11]'."
            ) from e
        self._signer = pkcs11.Signer(pkcs11_uri, module_paths)
        return self

    def use_pkcs11_certificate_signer(
        self,
        *,
        pkcs11_uri: str,
        signing_certificate: pathlib.Path,
        certificate_chain: Iterable[pathlib.Path],
        module_paths: Iterable[str] = frozenset(),
    ) -> Self:
        """Configures the signing to be performed using signing certificates.

        The signer in this configuration is changed to one that performs signing
        using cryptographic certificates.

        Args:
            pkcs11_uri: The PKCS #11 URI.
            signing_certificate: The path to the signing certificate.
            certificate_chain: Optional paths to other certificates to establish
              a chain of trust.
            module_paths: Optional list of paths of PKCS #11 modules.

        Return:
            The new signing configuration.
        """
        try:
            from model_signing._signing import sign_pkcs11 as pkcs11
        except ImportError as e:
            raise RuntimeError(
                "PKCS #11 functionality requires the 'pkcs11' extra. "
                "Install with 'pip install model-signing[pkcs11]'."
            ) from e

        self._signer = pkcs11.CertSigner(
            pkcs11_uri,
            signing_certificate,
            certificate_chain,
            module_paths=module_paths,
        )
        return self



# Copyright 2025 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Sigstore based signature, signers and verifiers, using protobuf.

The difference between this module and `sign_sigstore` is that here we use
Sigstore via the protobuf specs instead of `sigstore-python`. This is to enable
support for traditional signing and verification. These require additional data
to be stored in the sigstore bundle used for the signature but `sigstore-python`
validation does not allow those.
"""

import abc
import json
import pathlib
import sys
from typing import cast

from sigstore_models.bundle import v1 as bundle_pb
from typing_extensions import override

from model_signing._signing import signing


if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


# The media type to use when creating the protobuf based Sigstore bundle
_BUNDLE_MEDIA_TYPE: str = "application/vnd.dev.sigstore.bundle.v0.3+json"


def pae(raw_payload: bytes) -> bytes:
    """Generates the PAE encoding of statement from the payload.

    This is an internal of `sigstore_python`, but since in this module and
    classes derived from the signer and verifier defined here we cannot use
    `sigstore_python`, we have to reimplement this.

    See https://github.com/secure-systems-lab/dsse/blob/v1.0.0/protocol.md
    for details.

    Args:
        payload: The raw payload to encode.

    Returns:
        The encoded statement from the payload.
    """
    payload_type = signing._IN_TOTO_JSON_PAYLOAD_TYPE
    payload_type_length = len(payload_type)
    payload_length = len(raw_payload)
    pae_str = f"DSSEv1 {payload_type_length} {payload_type} {payload_length}"
    return b" ".join([pae_str.encode("utf-8"), raw_payload])


def pae_compat(raw_payload: bytes) -> bytes:
    """Generates the PAE encoding of statement from the payload.

    This is the same as `pae`, but using the version as defined in v0.2.0 of the
    `model_signing` library. Due to a bug in that implementation, signatures
    generated at that version have to be verified using this compat patch. The
    issue is that the raw payload, which is bytes, is added to a string and then
    encoded back as bytes, so we get additional escape characters included.

    Args:
        payload: The raw payload to encode.

    Returns:
        The encoded statement from the payload.
    """
    payload_type = signing._IN_TOTO_JSON_PAYLOAD_TYPE
    payload_type_length = len(payload_type)
    payload_length = len(raw_payload)
    # Notice bug here!
    pae_str = (
        f"DSSEV1 {payload_type_length} {payload_type} "
        f"{payload_length} {raw_payload}"
    )
    return pae_str.encode("utf-8")


class Signature(signing.Signature):
    """Sigstore signature support, wrapping around `bundle_pb.Bundle`."""

    def __init__(self, bundle: bundle_pb.Bundle):
        """Builds an instance of this signature.

        Args:
            bundle: the sigstore bundle (in `bundle_pb.Bundle` format).
        """
        self.bundle = bundle

    @override
    def write(self, path: pathlib.Path) -> None:
        path.write_text(self.bundle.to_json(), encoding="utf-8")

    @classmethod
    @override
    def read(cls, path: pathlib.Path) -> Self:
        content = path.read_text(encoding="utf-8")
        parsed_dict = json.loads(content)

        # adjust parsed_dict due to previous usage of protobufs
        if "tlogEntries" not in parsed_dict["verificationMaterial"]:
            parsed_dict["verificationMaterial"]["tlogEntries"] = []
        if "publicKey" in parsed_dict["verificationMaterial"]:
            if "hint" not in parsed_dict["verificationMaterial"]["publicKey"]:
                parsed_dict["verificationMaterial"]["publicKey"]["hint"] = None
            for k in ["rawBytes", "keyDetails"]:
                if k in parsed_dict["verificationMaterial"]["publicKey"]:
                    del parsed_dict["verificationMaterial"]["publicKey"][k]

        return cls(bundle_pb.Bundle.from_dict(parsed_dict))


class Signer(signing.Signer):
    """Signer for traditional signing.

    This is subclassed for each traditional signing method we support.
    """


class Verifier(signing.Verifier):
    """Verifier for traditional signature verification.

    This is subclassed for each traditional signing method we support.
    """

    @override
    def _verify_signed_content(
        self, signature: signing.Signature
    ) -> tuple[str, bytes]:
        # We are guaranteed to only use the local signature type
        signature = cast(Signature, signature)
        bundle = signature.bundle

        # Since the bundle is done via protobuf, check media type first
        if bundle.media_type != _BUNDLE_MEDIA_TYPE:
            raise ValueError(
                f"Invalid sigstore bundle, got media type {bundle.media_type} "
                f"but expected {_BUNDLE_MEDIA_TYPE}"
            )

        return self._verify_bundle(bundle)

    @abc.abstractmethod
    def _verify_bundle(self, bundle: bundle_pb.Bundle) -> tuple[str, bytes]:
        """Verifies the bundle to extract the payload type and payload.

        Since the bundle is generated via proto, we need to do more checks to
        replace what `verify_dsse` from `sigstore_python` does.
        """



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Machinery for computing digests for a single object.

The `Digest` object represents the algorithm used to summarize (hash) an object
(file, chunk of file, etc.) and the value of the digest. These are the values
that are stored in signature files.

To support multiple hashing formats, we define an abstract `HashEngine` class
which can be used in type annotations and is at the root of the hashing classes
hierarchy.

To support updating the hash of an object when more data is being added, we also
define a `Streaming` protocol which we then use to define an abstract
`StreamingHashEngine`.

These two types of hashing engines are used to hash any of the objects that we
can generate a signature over.
"""

import abc
import dataclasses
from typing import Protocol


@dataclasses.dataclass(frozen=True)
class Digest:
    """A digest computed by a `HashEngine`.

    Attributes:
        algorithm: The algorithm used to compute the digest. This could be a
          canonical name (e.g. "sha256" for SHA256) or a name that uniquely
          encodes the algorithm being used for the purposes of this library
          (e.g., "sha256-sharded-1024" for a digest produced by computing SHA256
          hashes of shards of 1024 bytes of the object).  This name can be used
          to autodetect the hashing configuration used during signing so that
          verification can compute a similar digest.
        digest_value: The value of the digest.
    """

    algorithm: str
    digest_value: bytes

    @property
    def digest_hex(self) -> str:
        """Hexadecimal, human readable, equivalent of `digest`."""
        return self.digest_value.hex()

    @property
    def digest_size(self) -> int:
        """The size, in bytes, of the digest."""
        return len(self.digest_value)


class HashEngine(metaclass=abc.ABCMeta):
    """Generic hash engine."""

    @abc.abstractmethod
    def compute(self) -> Digest:
        """Computes the digest of data passed to the engine."""

    @property
    @abc.abstractmethod
    def digest_name(self) -> str:
        """The canonical name of the algorithm used to compute the hash.

        Subclasses MUST use the `digest_name()` method to record all parameters
        that influence the hash output. For example, if a file is split into
        shards which are hashed separately and the final digest value is
        computed by aggregating these hashes, then the shard size must be given
        in the output string.

        This name gets transferred to the `algorithm` field of the `Digest`
        computed by the hashing engine.
        """

    @property
    @abc.abstractmethod
    def digest_size(self) -> int:
        """The size, in bytes, of the digests produced by the engine.

        This must return the same value as calling `digest_size` on the `Digest`
        object produced by the hashing engine.
        """


class Streaming(Protocol):
    """A protocol to support streaming data to `HashEngine` objects."""

    @abc.abstractmethod
    def update(self, data: bytes) -> None:
        """Appends additional bytes to the data to be hashed.

        Args:
            data: The new data that should be hashed.
        """

    @abc.abstractmethod
    def reset(self, data: bytes = b"") -> None:
        """Resets the data to be hashed to the passed argument.

        Args:
            data: Optional, initial data to hash.
        """


class StreamingHashEngine(Streaming, HashEngine):
    """A `HashEngine` that can stream data to be hashed."""



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Model serializers that operate at file shard level of granularity."""

from collections.abc import Callable, Iterable
import concurrent.futures
import itertools
import os
import pathlib
from typing import Optional

from typing_extensions import override

from model_signing import manifest
from model_signing._hashing import io
from model_signing._serialization import serialization


def _endpoints(step: int, end: int) -> Iterable[int]:
    """Yields numbers from `step` to `end` inclusive, spaced by `step`.

    Last value is always equal to `end`, even when `end` is not a multiple of
    `step`. There is always a value returned.

    Examples:
    ```python
    >>> list(_endpoints(2, 8))
    [2, 4, 6, 8]
    >>> list(_endpoints(2, 9))
    [2, 4, 6, 8, 9]
    >>> list(_endpoints(2, 2))
    [2]

    Yields:
        Values in the range, from `step` and up to `end`.
    """
    yield from range(step, end, step)
    yield end


class Serializer(serialization.Serializer):
    """Model serializer that produces a manifest recording every file shard.

    Traverses the model directory and creates digests for every file found,
    sharding the file in equal shards and computing the digests in parallel.
    """

    def __init__(
        self,
        sharded_hasher_factory: Callable[
            [pathlib.Path, int, int], io.ShardedFileHasher
        ],
        *,
        max_workers: Optional[int] = None,
        allow_symlinks: bool = False,
        ignore_paths: Iterable[pathlib.Path] = frozenset(),
    ):
        """Initializes an instance to serialize a model with this serializer.

        Args:
            sharded_hasher_factory: A callable to build the hash engine used to
              hash every shard of the files in the model. Because each shard is
              processed in parallel, every thread needs to call the factory to
              start hashing. The arguments are the file, and the endpoints of
              the shard.
            max_workers: Maximum number of workers to use in parallel. Default
              is to defer to the `concurrent.futures` library.
            allow_symlinks: Controls whether symbolic links are included. If a
              symlink is present but the flag is `False` (default) the
              serialization would raise an error.
            ignore_paths: The paths of files to ignore.
        """
        self._hasher_factory = sharded_hasher_factory
        self._max_workers = max_workers
        self._allow_symlinks = allow_symlinks
        self._ignore_paths = ignore_paths

        # Precompute some private values only once by using a mock file hasher.
        # None of the arguments used to build the hasher are used.
        hasher = sharded_hasher_factory(pathlib.Path(), 0, 1)
        self._shard_size = hasher.shard_size
        self._serialization_description = manifest._ShardSerialization(
            # Here we need the internal hasher name, not the mangled name.
            # This name is used when guessing the hashing configuration.
            hasher._content_hasher.digest_name,
            self._shard_size,
            self._allow_symlinks,
            self._ignore_paths,
        )

    def set_allow_symlinks(self, allow_symlinks: bool) -> None:
        """Set whether following symlinks is allowed."""
        self._allow_symlinks = allow_symlinks
        hasher = self._hasher_factory(pathlib.Path(), 0, 1)
        self._serialization_description = manifest._ShardSerialization(
            hasher._content_hasher.digest_name,
            self._shard_size,
            self._allow_symlinks,
            self._ignore_paths,
        )

    @override
    def serialize(
        self,
        model_path: pathlib.Path,
        *,
        ignore_paths: Iterable[pathlib.Path] = frozenset(),
        files_to_hash: Optional[Iterable[pathlib.Path]] = None,
    ) -> manifest.Manifest:
        """Serializes the model given by the `model_path` argument.

        Args:
            model_path: The path to the model.
            ignore_paths: The paths to ignore during serialization. If a
              provided path is a directory, all children of the directory are
              ignored.
            files_to_hash: Optional list of files to hash; ignore all others

        Returns:
            The model's serialized manifest.

        Raises:
            ValueError: The model contains a symbolic link, but the serializer
              was not initialized with `allow_symlinks=True`.
        """
        shards = []
        # TODO: github.com/sigstore/model-transparency/issues/200 - When
        # Python3.12 is the minimum supported version, the glob can be replaced
        # with `pathlib.Path.walk` for a clearer interface, and some speed
        # improvement.
        if files_to_hash is None:
            files_to_hash = itertools.chain(
                (model_path,), model_path.glob("**/*")
            )
        for path in files_to_hash:
            serialization.check_file_or_directory(
                path, allow_symlinks=self._allow_symlinks
            )
            if path.is_file() and not serialization.should_ignore(
                path, ignore_paths
            ):
                shards.extend(self._get_shards(path))

        manifest_items = []
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self._max_workers
        ) as tpe:
            futures = [
                tpe.submit(self._compute_hash, model_path, path, start, end)
                for path, start, end in shards
            ]
            for future in concurrent.futures.as_completed(futures):
                manifest_items.append(future.result())

        # Recreate serialization_description for new ignore_paths
        if ignore_paths:
            rel_ignore_paths = []
            for p in ignore_paths:
                rp = os.path.relpath(p, model_path)
                # rp may start with "../" if it is not relative to model_path
                if not rp.startswith("../"):
                    rel_ignore_paths.append(pathlib.Path(rp))

            hasher = self._hasher_factory(pathlib.Path(), 0, 1)
            self._serialization_description = manifest._ShardSerialization(
                hasher._content_hasher.digest_name,
                self._shard_size,
                self._allow_symlinks,
                frozenset(list(self._ignore_paths) + rel_ignore_paths),
            )

        model_name = model_path.name
        if not model_name or model_name == "..":
            model_name = os.path.basename(model_path.resolve())

        return manifest.Manifest(
            model_name, manifest_items, self._serialization_description
        )

    def _get_shards(
        self, path: pathlib.Path
    ) -> list[tuple[pathlib.Path, int, int]]:
        """Determines the shards of a given file path."""
        shards = []
        path_size = path.stat().st_size
        if path_size > 0:
            start = 0
            for end in _endpoints(self._shard_size, path_size):
                shards.append((path, start, end))
                start = end
        return shards

    def _compute_hash(
        self, model_path: pathlib.Path, path: pathlib.Path, start: int, end: int
    ) -> manifest.ShardedFileManifestItem:
        """Produces the manifest item of the file given by `path`.

        Args:
            model_path: The path to the model.
            path: Path to the file in the model, that is currently transformed
              to a manifest item.
            start: The start offset of the shard (included).
            end: The end offset of the shard (not included).

        Returns:
            The itemized manifest.
        """
        relative_path = path.relative_to(model_path)
        digest = self._hasher_factory(path, start, end).compute()
        return manifest.ShardedFileManifestItem(
            path=relative_path, digest=digest, start=start, end=end
        )



# Copyright 2024 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Machinery for serializing ML models."""

import abc
from collections.abc import Iterable
import pathlib

from model_signing import manifest


def check_file_or_directory(
    path: pathlib.Path, *, allow_symlinks: bool = False
) -> None:
    """Checks that the given path is either a file or a directory.

    There is no support for sockets, pipes, or any other operating system
    concept abstracted as a file.

    Furthermore, this would raise if the path is a broken symlink, if it doesn't
    exists or if there are permission errors.

    Args:
        path: The path to check.
        allow_symlinks: Controls whether symbolic links are included. If a
          symlink is present but the flag is `False` (default) the
          serialization would raise an error.

    Raises:
        ValueError: The path is neither a file or a directory, or the path
          is a symlink and `allow_symlinks` is false.
    """
    if not allow_symlinks and path.is_symlink():
        raise ValueError(
            f"Cannot use '{path}' because it is a symlink. This"
            " behavior can be changed with `allow_symlinks`."
        )
    if not (path.is_file() or path.is_dir()):
        raise ValueError(
            f"Cannot use '{path}' as file or directory. It could be a"
            " special file, it could be missing, or there might be a"
            " permission issue."
        )


def should_ignore(
    path: pathlib.Path, ignore_paths: Iterable[pathlib.Path]
) -> bool:
    """Determines if the provided path should be ignored during serialization.

    Args:
        path: The path to check.
        ignore_paths: The paths to ignore while serializing a model.

    Returns:
        Whether or not the provided path should be ignored.
    """
    return any(path.is_relative_to(ignore_path) for ignore_path in ignore_paths)


class Serializer(metaclass=abc.ABCMeta):
    """Generic ML model format serializer."""

    @abc.abstractmethod
    def serialize(
        self,
        model_path: pathlib.Path,
        *,
        ignore_paths: Iterable[pathlib.Path] = frozenset(),
    ) -> manifest.Manifest:
        """Serializes the model given by the `model_path` argument.

        Args:
            model_path: The path to the model.
            ignore_paths: The paths to ignore during serialization. If a
              provided path is a directory, all children of the directory are
              ignored.

        Returns:
            The model's serialized manifest.
        """



# Copyright 2025 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The main entry-point for the model_signing package."""

from collections.abc import Iterable, Sequence
import contextlib
import logging
import pathlib
import sys
from typing import Optional

import click

import model_signing


class NoOpTracer:
    def start_as_current_span(self, name):
        @contextlib.contextmanager
        def noop_context():
            class NoOpSpan:
                def set_attribute(self, key, value):
                    pass

            yield NoOpSpan()

        return noop_context()


# Global tracer variable, we will initialized this within the main() function
tracer = None


# Decorator for the commonly used argument for the model path.
_model_path_argument = click.argument(
    "model_path", type=pathlib.Path, metavar="MODEL_PATH"
)


# Decorator for the commonly used option to set the signature path when signing.
_write_signature_option = click.option(
    "--signature",
    type=pathlib.Path,
    metavar="SIGNATURE_PATH",
    default=pathlib.Path("model.sig"),
    help="Location of the signature file to generate. Defaults to `model.sig`.",
)


# Decorator for the commonly used option for the signature to verify.
_read_signature_option = click.option(
    "--signature",
    type=pathlib.Path,
    metavar="SIGNATURE_PATH",
    required=True,
    help="Location of the signature file to verify.",
)

# Decorator for the commonly used option for the custom trust configuration.
_trust_config_option = click.option(
    "--trust_config",
    type=pathlib.Path,
    metavar="TRUST_CONFIG_PATH",
    help="The client trust configuration to use",
)

# Decorator for the commonly used option to ignore certain paths
_ignore_paths_option = click.option(
    "--ignore-paths",
    type=pathlib.Path,
    metavar="IGNORE_PATHS",
    multiple=True,
    help="File paths to ignore when signing or verifying.",
)

# Decorator for the commonly used option to ignore git-related paths
_ignore_git_paths_option = click.option(
    "--ignore-git-paths/--no-ignore-git-paths",
    type=bool,
    default=True,
    show_default=True,
    help="Ignore git-related files when signing or verifying.",
)

# Decorator for the commonly used option to ignore all unsigned files
_ignore_unsigned_files_option = click.option(
    "--ignore_unsigned_files/--no-ignore_unsigned_files",
    type=bool,
    show_default=True,
    help="Ignore all files that were not originally signed.",
)

# Decorator for the commonly used option to set the path to the private key
# (when using non-Sigstore PKI).
_private_key_option = click.option(
    "--private_key",
    type=pathlib.Path,
    metavar="PRIVATE_KEY",
    required=True,
    help="Path to the private key, as a PEM-encoded file.",
)

# Decorator for the commonly used option to set a PKCS #11 URI
_pkcs11_uri_option = click.option(
    "--pkcs11_uri",
    type=str,
    metavar="PKCS11_URI",
    required=True,
    help="PKCS #11 URI of the private key.",
)

# Decorator for the commonly used option to pass a certificate chain to
# establish root of trust (when signing or verifying using certificates).
_certificate_root_of_trust_option = click.option(
    "--certificate_chain",
    type=pathlib.Path,
    metavar="CERTIFICATE_PATH",
    multiple=True,
    help="Path to certificate chain of trust.",
)


# Decorator for the commonly used option to use Sigstore's staging instance.
_sigstore_staging_option = click.option(
    "--use_staging",
    type=bool,
    is_flag=True,
    help="Use Sigstore's staging instance.",
)

# Decorator for the commonly used option to pass the signing key's certificate
_signing_certificate_option = click.option(
    "--signing_certificate",
    type=pathlib.Path,
    metavar="CERTIFICATE_PATH",
    required=True,
    help="Path to the signing certificate, as a PEM-encoded file.",
)

# Decorator for the commonly used option to allow symlinks
_allow_symlinks_option = click.option(
    "--allow_symlinks",
    is_flag=True,
    help="Whether to allow following symlinks when signing or verifying files.",
)


def _resolve_ignore_paths(
    model_path: pathlib.Path, paths: Iterable[pathlib.Path]
) -> list[pathlib.Path]:
    model_root = model_path.resolve()
    cwd = pathlib.Path.cwd()
    resolved_paths = []
    for p in paths:
        candidate = (p if p.is_absolute() else (cwd / p)).resolve()
        try:
            resolved_paths.append(candidate.relative_to(model_root))
        except ValueError:
            continue
    return resolved_paths


class _PKICmdGroup(click.Group):
    """A custom group to configure the supported PKI methods."""

    _supported_modes = [
        "sigstore",
        "key",
        "certificate",
        "pkcs11-key",
        "pkcs11-certificate",
    ]

    def get_command(
        self, ctx: click.Context, cmd_name: str
    ) -> Optional[click.Command]:
        """Retrieves a command with a given name.

        We use this to make Sigstore signing be the default, if it is missing.
        """
        if cmd_name in self._supported_modes:
            return super().get_command(ctx, cmd_name)
        return super().get_command(ctx, "sigstore")

    def resolve_command(
        self, ctx: click.Context, args: Sequence[str]
    ) -> tuple[Optional[str], Optional[click.Command], Iterable[str]]:
        """Resolves a command and its arguments.

        We use this to make Sigstore signing be the default and correctly alter
        the arguments. We are guaranteed that `args` has at least one element
        (otherwise the help menu would be printed). This argument should be the
        subcommand and would be removed as a result of this function, in
        general.

        However, if the first argument does not resolve to a supported PKI
        method, then we inject "sigstore" as the subcommand (in `get_command`).
        All that is left to do is to pass all `args` to the subcommand, without
        removing anything.
        """
        if args[0] in self._supported_modes:
            return super().resolve_command(ctx, args)
        _, cmd, _ = super().resolve_command(ctx, args)
        return cmd.name, cmd, args


@click.group(
    context_settings=dict(help_option_names=["-h", "--help"]),
    epilog=(
        "Check https://sigstore.github.io/model-transparency for "
        "documentation and more details."
    ),
)
@click.version_option(model_signing.__version__, "--version")
@click.option(
    "--log-level",
    type=click.Choice(
        ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False
    ),
    default="INFO",
    show_default=True,
    metavar="LEVEL",
    help="Set the logging level. This can also be set via the "
    "MODEL_SIGNING_LOG_LEVEL env var.",
)
def main(log_level: str) -> None:
    """ML model signing and verification.

    Use each subcommand's `--help` option for details on each mode.
    """
    global tracer

    logging.basicConfig(
        format="%(message)s", level=getattr(logging, log_level.upper())
    )

    try:
        from opentelemetry import trace  # type: ignore[import-error]
        from opentelemetry.instrumentation import (
            auto_instrumentation,  # type: ignore[import-error]
        )

        auto_instrumentation.initialize()
        tracer = trace.get_tracer(__name__)
    except ImportError:
        logging.debug("OpenTelemetry not installed. Tracing is disabled.")
        tracer = NoOpTracer()
    except Exception as e:
        logging.error(
            f"Failed to initialize OpenTelemetry auto instrumentation: {e}"
        )
        sys.exit(1)


@main.group(name="sign", subcommand_metavar="PKI_METHOD", cls=_PKICmdGroup)
def _sign() -> None:
    """Sign models.

    Produces a cryptographic signature (in the form of a Sigstore bundle) for a
    model. We support any model format, either as a single file or as a
    directory.

    We support multiple PKI methods, specified as subcommands. By default, the
    signature is generated via Sigstore (as if invoking `sigstore` subcommand).

    Use each subcommand's `--help` option for details on each mode.
    """


@_sign.command(name="sigstore")
@_model_path_argument
@_ignore_paths_option
@_ignore_git_paths_option
@_allow_symlinks_option
@_write_signature_option
@_sigstore_staging_option
@_trust_config_option
@click.option(
    "--use_ambient_credentials",
    type=bool,
    is_flag=True,
    help="Use credentials from ambient environment.",
)
@click.option(
    "--identity_token",
    type=str,
    metavar="TOKEN",
    help=(
        "Fixed OIDC identity token to use instead of obtaining credentials "
        "from OIDC flow or from the environment."
    ),
)
@click.option(
    "--oauth_force_oob",
    is_flag=True,
    default=False,
    help=(
        "Force an out-of-band OAuth flow and do not automatically start "
        "the default web browser."
    ),
)
@click.option(
    "--client_id",
    type=str,
    metavar="ID",
    help="The custom OpenID Connect client ID to use during OAuth2",
)
@click.option(
    "--client_secret",
    type=str,
    metavar="SECRET",
    help="The custom OpenID Connect client secret to use during OAuth2",
)
def _sign_sigstore(
    model_path: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    allow_symlinks: bool,
    signature: pathlib.Path,
    use_ambient_credentials: bool,
    use_staging: bool,
    oauth_force_oob: bool,
    identity_token: Optional[str] = None,
    client_id: Optional[str] = None,
    client_secret: Optional[str] = None,
    trust_config: Optional[pathlib.Path] = None,
) -> None:
    """Sign using Sigstore (DEFAULT signing method).

    Signing the model at MODEL_PATH, produces the signature at SIGNATURE_PATH
    (as per `--signature` option). Files in IGNORE_PATHS are not part of the
    signature.

    If using Sigstore, we need to provision an OIDC token. In general, this is
    taken from an interactive OIDC flow, but ambient credentials could be used
    to use workload identity tokens (e.g., when running in GitHub actions).
    Alternatively, a constant identity token can be provided via
    `--identity_token`.

    Sigstore allows users to use a staging instance for test-only signatures.
    Passing the `--use_staging` flag would use that instance instead of the
    production one.

    Additionally, you can specify a custom trust configuration JSON file using
    the `--trust_config` flag. This allows you to fully customize the PKI
    (Private Key Infrastructure) used in the signing process. By providing a
    `--trust_config`, you can define your own transparency logs, certificate
    authorities, and other trust settings, enabling full control over the
    trust model, including which PKI to use for signature verification.

    If `--trust_config` is not provided, the default Sigstore instance is
    used, which is pre-configured with Sigstore’s own trusted transparency
    logs and certificate authorities. This provides a ready-to-use default
    trust model for most use cases but may not be suitable for custom or
    highly regulated environments.
    """
    with tracer.start_as_current_span("Sign") as span:
        span.set_attribute("sigstore.sign_method", "sigstore")
        span.set_attribute("sigstore.model_path", str(model_path))
        span.set_attribute("sigstore.signature", str(signature))
        span.set_attribute(
            "sigstore.use_ambient_credentials", use_ambient_credentials
        )
        span.set_attribute("sigstore.use_staging", use_staging)
        try:
            ignored = _resolve_ignore_paths(
                model_path, list(ignore_paths) + [signature]
            )
            model_signing.signing.Config().use_sigstore_signer(
                use_ambient_credentials=use_ambient_credentials,
                use_staging=use_staging,
                identity_token=identity_token,
                force_oob=oauth_force_oob,
                client_id=client_id,
                client_secret=client_secret,
                trust_config=trust_config,
            ).set_hashing_config(
                model_signing.hashing.Config()
                .set_ignored_paths(
                    paths=ignored, ignore_git_paths=ignore_git_paths
                )
                .set_allow_symlinks(allow_symlinks)
            ).sign(model_path, signature)
        except Exception as err:
            click.echo(f"Signing failed with error: {err}", err=True)
            sys.exit(1)

        click.echo("Signing succeeded")


@_sign.command(name="key")
@_model_path_argument
@_ignore_paths_option
@_ignore_git_paths_option
@_allow_symlinks_option
@_write_signature_option
@_private_key_option
@click.option(
    "--password",
    type=str,
    metavar="PASSWORD",
    help="Password for the key encryption, if any",
)
def _sign_private_key(
    model_path: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    allow_symlinks: bool,
    signature: pathlib.Path,
    private_key: pathlib.Path,
    password: Optional[str] = None,
) -> None:
    """Sign using a private key (paired with a public one).

    Signing the model at MODEL_PATH, produces the signature at SIGNATURE_PATH
    (as per `--signature` option). Files in IGNORE_PATHS are not part of the
    signature.

    Traditionally, signing could be achieved by using a public/private key pair.
    Pass the signing key using `--private_key`.

    Note that this method does not provide a way to tie to the identity of the
    signer, outside of pairing the keys. Also note that we don't offer key
    management protocols.
    """
    try:
        ignored = _resolve_ignore_paths(
            model_path, list(ignore_paths) + [signature]
        )
        model_signing.signing.Config().use_elliptic_key_signer(
            private_key=private_key, password=password
        ).set_hashing_config(
            model_signing.hashing.Config()
            .set_ignored_paths(paths=ignored, ignore_git_paths=ignore_git_paths)
            .set_allow_symlinks(allow_symlinks)
        ).sign(model_path, signature)
    except Exception as err:
        click.echo(f"Signing failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Signing succeeded")


@_sign.command(name="pkcs11-key")
@_model_path_argument
@_ignore_paths_option
@_ignore_git_paths_option
@_allow_symlinks_option
@_write_signature_option
@_pkcs11_uri_option
def _sign_pkcs11_key(
    model_path: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    allow_symlinks: bool,
    signature: pathlib.Path,
    pkcs11_uri: str,
) -> None:
    """Sign using a private key using a PKCS #11 URI.

    Signing the model at MODEL_PATH, produces the signature at SIGNATURE_PATH
    (as per `--signature` option). Files in IGNORE_PATHS are not part of the
    signature.

    Traditionally, signing could be achieved by using a public/private key pair.
    Pass the PKCS #11 URI of the signing key using `--pkcs11_uri`.

    Note that this method does not provide a way to tie to the identity of the
    signer, outside of pairing the keys. Also note that we don't offer key
    management protocols.
    """
    try:
        ignored = _resolve_ignore_paths(
            model_path, list(ignore_paths) + [signature]
        )
        model_signing.signing.Config().use_pkcs11_signer(
            pkcs11_uri=pkcs11_uri
        ).set_hashing_config(
            model_signing.hashing.Config()
            .set_ignored_paths(paths=ignored, ignore_git_paths=ignore_git_paths)
            .set_allow_symlinks(allow_symlinks)
        ).sign(model_path, signature)
    except Exception as err:
        click.echo(f"Signing failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Signing succeeded")


@_sign.command(name="certificate")
@_model_path_argument
@_ignore_paths_option
@_ignore_git_paths_option
@_allow_symlinks_option
@_write_signature_option
@_private_key_option
@_signing_certificate_option
@_certificate_root_of_trust_option
def _sign_certificate(
    model_path: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    allow_symlinks: bool,
    signature: pathlib.Path,
    private_key: pathlib.Path,
    signing_certificate: pathlib.Path,
    certificate_chain: Iterable[pathlib.Path],
) -> None:
    """Sign using a certificate.

    Signing the model at MODEL_PATH, produces the signature at SIGNATURE_PATH
    (as per `--signature` option). Files in IGNORE_PATHS are not part of the
    signature.

    Traditionally, signing can be achieved by using keys from a certificate.
    The certificate can also provide the identity of the signer, making this
    method more informative than just using a public/private key pair for
    signing.  Pass the private signing key using `--private_key` and signing
    certificate via `--signing_certificate`. Optionally, pass a certificate
    chain via `--certificate_chain` to establish root of trust (this option can
    be repeated as needed, or all cerificates could be placed in a single file).

    Note that we don't offer certificate and key management protocols.
    """
    try:
        ignored = _resolve_ignore_paths(
            model_path, list(ignore_paths) + [signature]
        )
        model_signing.signing.Config().use_certificate_signer(
            private_key=private_key,
            signing_certificate=signing_certificate,
            certificate_chain=certificate_chain,
        ).set_hashing_config(
            model_signing.hashing.Config()
            .set_ignored_paths(paths=ignored, ignore_git_paths=ignore_git_paths)
            .set_allow_symlinks(allow_symlinks)
        ).sign(model_path, signature)
    except Exception as err:
        click.echo(f"Signing failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Signing succeeded")


@_sign.command(name="pkcs11-certificate")
@_model_path_argument
@_ignore_paths_option
@_ignore_git_paths_option
@_allow_symlinks_option
@_write_signature_option
@_pkcs11_uri_option
@_signing_certificate_option
@_certificate_root_of_trust_option
def _sign_pkcs11_certificate(
    model_path: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    allow_symlinks: bool,
    signature: pathlib.Path,
    pkcs11_uri: str,
    signing_certificate: pathlib.Path,
    certificate_chain: Iterable[pathlib.Path],
) -> None:
    """Sign using a certificate.

    Signing the model at MODEL_PATH, produces the signature at SIGNATURE_PATH
    (as per `--signature` option). Files in IGNORE_PATHS are not part of the
    signature.

    Traditionally, signing can be achieved by using keys from a certificate.
    The certificate can also provide the identity of the signer, making this
    method more informative than just using a public/private key pair for
    signing. Pass the PKCS #11 URI of the private signing key using
    `--pkcs11_uri` and then signing certificate via `--signing_certificate`.
    Optionally, pass a certificate chain via `--certificate_chain` to establish
    root of trust (this option can be repeated as needed, or all cerificates
    could be placed in a single file).

    Note that we don't offer certificate and key management protocols.
    """
    try:
        ignored = _resolve_ignore_paths(
            model_path, list(ignore_paths) + [signature]
        )
        model_signing.signing.Config().use_pkcs11_certificate_signer(
            pkcs11_uri=pkcs11_uri,
            signing_certificate=signing_certificate,
            certificate_chain=certificate_chain,
        ).set_hashing_config(
            model_signing.hashing.Config()
            .set_ignored_paths(paths=ignored, ignore_git_paths=ignore_git_paths)
            .set_allow_symlinks(allow_symlinks)
        ).sign(model_path, signature)
    except Exception as err:
        click.echo(f"Signing failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Signing succeeded")


@main.group(name="verify", subcommand_metavar="PKI_METHOD", cls=_PKICmdGroup)
def _verify() -> None:
    """Verify models.

    Given a model and a cryptographic signature (in the form of a Sigstore
    bundle) for the model, this call checks that the model matches the
    signature, that the model has not been tampered with. We support any model
    format, either as a signle file or as a directory.

    We support multiple PKI methods, specified as subcommands. By default, the
    signature is assumed to be generated via Sigstore (as if invoking `sigstore`
    subcommand).

    To enable verification with custom PKI configurations, use the
    `--trust_config` option. This allows you to specify your own set of trusted
    public keys, transparency logs, and certificate authorities for verifying
    the signature. If not provided, the default Sigstore instance and its
    associated public keys, logs, and authorities are used.

    Use each subcommand's `--help` option for details on each mode.
    """


@_verify.command(name="sigstore")
@_model_path_argument
@_read_signature_option
@_ignore_paths_option
@_ignore_git_paths_option
@_allow_symlinks_option
@_sigstore_staging_option
@_trust_config_option
@click.option(
    "--identity",
    type=str,
    metavar="IDENTITY",
    required=True,
    help="The expected identity of the signer (e.g., name@example.com).",
)
@click.option(
    "--identity_provider",
    type=str,
    metavar="IDENTITY_PROVIDER",
    required=True,
    help="The expected identity provider (e.g., https://accounts.example.com).",
)
@_ignore_unsigned_files_option
def _verify_sigstore(
    model_path: pathlib.Path,
    signature: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    allow_symlinks: bool,
    identity: str,
    identity_provider: str,
    use_staging: bool,
    ignore_unsigned_files: bool,
    trust_config: Optional[pathlib.Path] = None,
) -> None:
    """Verify using Sigstore (DEFAULT verification method).

    Verifies the integrity of model at MODEL_PATH, according to signature from
    SIGNATURE_PATH (given via `--signature` option). Files in IGNORE_PATHS are
    ignored.

    For Sigstore, we also need to provide an expected identity and identity
    provider for the signature. If these don't match what is provided in the
    signature, verification would fail.
    """
    with tracer.start_as_current_span("Verify") as span:
        span.set_attribute("sigstore.method", "sigstore")
        span.set_attribute("sigstore.model_path", str(model_path))
        span.set_attribute("sigstore.signature", str(signature))
        span.set_attribute("sigstore.identity", identity)
        span.set_attribute("sigstore.oidc_issuer", identity_provider)
        span.set_attribute("sigstore.use_staging", use_staging)
        try:
            ignored = _resolve_ignore_paths(
                model_path, list(ignore_paths) + [signature]
            )
            model_signing.verifying.Config().use_sigstore_verifier(
                identity=identity,
                oidc_issuer=identity_provider,
                use_staging=use_staging,
                trust_config=trust_config,
            ).set_hashing_config(
                model_signing.hashing.Config()
                .set_ignored_paths(
                    paths=ignored, ignore_git_paths=ignore_git_paths
                )
                .set_allow_symlinks(allow_symlinks)
            ).set_ignore_unsigned_files(ignore_unsigned_files).verify(
                model_path, signature
            )
        except Exception as err:
            click.echo(f"Verification failed with error: {err}", err=True)
            sys.exit(1)

        click.echo("Verification succeeded")


@_verify.command(name="key")
@_model_path_argument
@_read_signature_option
@_ignore_paths_option
@_ignore_git_paths_option
@_allow_symlinks_option
@click.option(
    "--public_key",
    type=pathlib.Path,
    metavar="PUBLIC_KEY",
    required=True,
    help="Path to the public key used for verification.",
)
@_ignore_unsigned_files_option
def _verify_private_key(
    model_path: pathlib.Path,
    signature: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    allow_symlinks: bool,
    public_key: pathlib.Path,
    ignore_unsigned_files: bool,
) -> None:
    """Verity using a public key (paired with a private one).

    Verifies the integrity of model at MODEL_PATH, according to signature from
    SIGNATURE_PATH (given via `--signature` option). Files in IGNORE_PATHS are
    ignored.

    The public key provided via `--public_key` must have been paired with the
    private key used when generating the signature.

    Note that this method does not provide a way to tie to the identity of the
    signer, outside of pairing the keys. Also note that we don't offer key
    management protocols.
    """
    try:
        ignored = _resolve_ignore_paths(
            model_path, list(ignore_paths) + [signature]
        )
        model_signing.verifying.Config().use_elliptic_key_verifier(
            public_key=public_key
        ).set_hashing_config(
            model_signing.hashing.Config()
            .set_ignored_paths(paths=ignored, ignore_git_paths=ignore_git_paths)
            .set_allow_symlinks(allow_symlinks)
        ).set_ignore_unsigned_files(ignore_unsigned_files).verify(
            model_path, signature
        )
    except Exception as err:
        click.echo(f"Verification failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Verification succeeded")


@_verify.command(name="certificate")
@_model_path_argument
@_read_signature_option
@_ignore_paths_option
@_ignore_git_paths_option
@_allow_symlinks_option
@_certificate_root_of_trust_option
@click.option(
    "--log_fingerprints",
    type=bool,
    is_flag=True,
    default=False,
    show_default=True,
    help="Log SHA256 fingerprints of all certificates.",
)
@_ignore_unsigned_files_option
def _verify_certificate(
    model_path: pathlib.Path,
    signature: pathlib.Path,
    ignore_paths: Iterable[pathlib.Path],
    ignore_git_paths: bool,
    allow_symlinks: bool,
    certificate_chain: Iterable[pathlib.Path],
    log_fingerprints: bool,
    ignore_unsigned_files: bool,
) -> None:
    """Verify using a certificate.

    Verifies the integrity of model at MODEL_PATH, according to signature from
    SIGNATURE_PATH (given via `--signature` option). Files in IGNORE_PATHS are
    ignored.

    The signing certificate is encoded in the signature, as part of the Sigstore
    bundle. To verify the root of trust, pass additional certificates in the
    certificate chain, using `--certificate_chain` (this option can be repeated
    as needed, or all certificates could be placed in a single file).

    Note that we don't offer certificate and key management protocols.
    """
    if log_fingerprints:
        logging.basicConfig(format="%(message)s", level=logging.INFO)

    try:
        ignored = _resolve_ignore_paths(
            model_path, list(ignore_paths) + [signature]
        )
        model_signing.verifying.Config().use_certificate_verifier(
            certificate_chain=certificate_chain,
            log_fingerprints=log_fingerprints,
        ).set_hashing_config(
            model_signing.hashing.Config()
            .set_ignored_paths(paths=ignored, ignore_git_paths=ignore_git_paths)
            .set_allow_symlinks(allow_symlinks)
        ).set_ignore_unsigned_files(ignore_unsigned_files).verify(
            model_path, signature
        )
    except Exception as err:
        click.echo(f"Verification failed with error: {err}", err=True)
        sys.exit(1)

    click.echo("Verification succeeded")



# Copyright 2025 The Sigstore Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from collections.abc import Iterable
import hashlib
import pathlib
from typing import Optional

from asn1crypto.algos import DSASignature
from asn1crypto.core import OctetString
from asn1crypto.keys import ECDomainParameters
from asn1crypto.keys import PublicKeyInfo
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ec
from google.protobuf import json_format
import PyKCS11
from sigstore_models import intoto as intoto_pb
from sigstore_models.bundle import v1 as bundle_pb
from sigstore_models.common import v1 as common_pb
from typing_extensions import override

from model_signing._signing import sign_ec_key as ec_key
from model_signing._signing import sign_sigstore_pb as sigstore_pb
from model_signing._signing import signing
from model_signing._signing.pkcs11uri import Pkcs11URI


MODULE_PATHS: Iterable[str] = [
    "/usr/lib64/pkcs11/",  # Fedora, RHEL, openSUSE
    "/usr/lib/pkcs11/",  # Fedora 32 bit, ArchLinux
]


def _check_supported_ec_key(public_key: ec.EllipticCurvePublicKey):
    """Checks if the elliptic curve key is supported by our package.

    We only support a family of curves, trying to match those specified by
    Sigstore's protobuf specs.
    See https://github.com/sigstore/model-transparency/issues/385.

    Args:
        public_key: The public key to check. Can be obtained from a private key.

    Raises:
        ValueError: The key is not supported.
    """
    curve = public_key.curve.name
    if curve not in ["secp256r1", "secp384r1", "secp521r1"]:
        raise ValueError(f"Unsupported key for curve '{curve}'")


def encode_ec_public_key(public_key: PyKCS11.CK_OBJECT_HANDLE) -> PublicKeyInfo:
    ec_params, ec_point = public_key.session.getAttributeValue(
        public_key, [PyKCS11.CKA_EC_PARAMS, PyKCS11.CKA_EC_POINT]
    )
    return PublicKeyInfo(
        {
            "algorithm": {
                "algorithm": "ec",
                "parameters": ECDomainParameters.load(bytes(ec_params)),
            },
            "public_key": bytes(OctetString.load(bytes(ec_point))),
        }
    ).dump()


class Signer(sigstore_pb.Signer):
    """Signer using PKCS #11 URIs with elliptic curves keys."""

    def __init__(
        self, pkcs11_uri: str, module_paths: Iterable[str] = frozenset()
    ):
        self.session = None

        self.pkcs11_uri = Pkcs11URI()
        self.pkcs11_uri.parse(pkcs11_uri)

        if len(list(module_paths)) == 0:
            module_paths = MODULE_PATHS

        # To support module-name set a few standard paths
        self.pkcs11_uri.set_module_directories(module_paths)
        self.pkcs11_uri.set_allow_any_module(True)

        self.session, self.lib = self.pkcs11_uri.login()

        keyid, label = self.pkcs11_uri.get_keyid_and_label()

        self._private_key = self.find_object(
            PyKCS11.CKO_PRIVATE_KEY, label=label, id=keyid
        )
        public_key = self.find_object(
            PyKCS11.CKO_PUBLIC_KEY, label=label, id=keyid
        )
        pub_der = encode_ec_public_key(public_key)

        # _public_key is a python cryptography key now
        self._public_key = serialization.load_der_public_key(pub_der)
        _check_supported_ec_key(self._public_key)

    def __del__(self):
        if self.session:
            try:
                self.session.closeSession()
            finally:
                pass

    def public_key(self):
        """Get the python cryptography public key."""
        return self._public_key

    def find_object(
        self, clas: int, label: Optional[str], id: Optional[bytes]
    ) -> PyKCS11.CK_OBJECT_HANDLE:
        """Find an object given its class and optional label and id."""
        if label is None and id is None:
            raise ValueError(
                "Missing search criteria for object: either label or id must "
                "be provided in URI"
            )

        msg = ""
        if label is not None:
            msg = f"label {label}"
        if id is not None:
            if len(msg):
                msg += " and "
            msg += f"id {id}"

        cka_id = None
        if id is not None:
            cka_id = tuple([x for x in id])

        for obj in self.session.findObjects([(PyKCS11.CKA_CLASS, clas)]):
            obj_label, obj_id = obj.session.getAttributeValue(
                obj, [PyKCS11.CKA_LABEL, PyKCS11.CKA_ID]
            )
            if label is not None and label != obj_label:
                continue
            if cka_id is not None and cka_id != obj_id:
                continue
            return obj
        raise ValueError(f"Could not find any object with {msg}")

    @override
    def sign(self, payload: signing.Payload) -> signing.Signature:
        raw_payload = json_format.MessageToJson(payload.statement.pb).encode(
            "utf-8"
        )

        hash_alg = ec_key.get_ec_key_hash(self._public_key)

        hash = hashes.Hash(hash_alg)
        hash.update(sigstore_pb.pae(raw_payload))
        digest = hash.finalize()

        rs_sig = self.session.sign(
            self._private_key,
            digest,
            mecha=PyKCS11.Mechanism(PyKCS11.CKM_ECDSA, None),
        )
        # Convert plain r & s signature values to ASN.1
        sig = DSASignature.from_p1363(rs_sig).dump()

        raw_signature = intoto_pb.Signature(sig=base64.b64encode(sig), keyid="")

        envelope = intoto_pb.Envelope(
            payload=base64.b64encode(raw_payload),
            payload_type=signing._IN_TOTO_JSON_PAYLOAD_TYPE,
            signatures=[raw_signature],
        )

        return sigstore_pb.Signature(
            bundle_pb.Bundle(
                media_type=sigstore_pb._BUNDLE_MEDIA_TYPE,
                verification_material=self._get_verification_material(),
                dsse_envelope=envelope,
            )
        )

    def _get_verification_material(self) -> bundle_pb.VerificationMaterial:
        """Returns the verification material to include in the bundle."""
        public_key = self._public_key

        raw_bytes = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo,
        )

        hash_bytes = hashlib.sha256(raw_bytes).digest().hex()

        return bundle_pb.VerificationMaterial(
            public_key=common_pb.PublicKeyIdentifier(hint=hash_bytes),
            tlog_entries=[],
        )


class CertSigner(Signer):
    """Signer using certificates."""

    def __init__(
        self,
        pkcs11_uri: str,
        signing_certificate_path: pathlib.Path,
        certificate_chain_paths: Iterable[pathlib.Path],
        module_paths: Iterable[str] = frozenset(),
    ):
        """Initializes the signer with the key, certificate and trust chain.

        Args:
            pkcs11_uri: The PKCS #11 URI.
            signing_certificate_path: The path to the signing certificate.
            certificate_chain_paths: Paths to other certificates used to
              establish chain of trust.

        Raises:
            ValueError: Signing certificate's public key does not match the
              private key's public pair.
        """
        super().__init__(pkcs11_uri, module_paths=module_paths)
        self._signing_certificate = x509.load_pem_x509_certificate(
            signing_certificate_path.read_bytes()
        )

        public_key_from_key = self._public_key
        public_key_from_certificate = self._signing_certificate.public_key()
        if public_key_from_key != public_key_from_certificate:
            raise ValueError(
                "The public key from the certificate does not match "
                "the public key paired with the private key"
            )

        self._trust_chain = x509.load_pem_x509_certificates(
            b"".join([path.read_bytes() for path in certificate_chain_paths])
        )

    @override
    def _get_verification_material(self) -> bundle_pb.VerificationMaterial:
        def _to_protobuf_certificate(certificate):
            return common_pb.X509Certificate(
                raw_bytes=certificate.public_bytes(
                    encoding=serialization.Encoding.DER
                )
            )

        chain = [_to_protobuf_certificate(self._signing_certificate)]
        chain.extend(
            [
                _to_protobuf_certificate(certificate)
                for certificate in self._trust_chain
            ]
        )

        return bundle_pb.VerificationMaterial(
            x509_certificate_chain=common_pb.X509CertificateChain(
                certificates=chain
            )
        )



# Copyright (c) 2025, IBM CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functionality to parse PKCS #11 URIs and access its components."""

from collections.abc import Iterable
import glob
import os
from os import stat
import re
from stat import S_ISREG
from typing import Optional
from urllib import parse
from urllib.parse import urlparse

import PyKCS11


def escape_all(s: str) -> str:
    """PCT-escape all characters in the given string; used primarily for Id."""
    res = ""
    for c in s:
        o = ord(c) & 0xFF
        res += f"%{o:02X}"

    return res


def escape(s: str, is_path: bool) -> str:
    """PCT-escape the given string considering path or query attribute."""
    res = ""
    for c in s:
        if (
            (
                (c >= "a" and c <= "z")
                or (c >= "A" and c <= "Z")
                or (c >= "0" and c <= "9")
            )
            or (is_path and c == "&")
            or (not is_path and (c in ["/", "?", "|"]))
        ):
            res += c
        else:
            if c in ["-", ".", "_", "~"] or c in [
                ":",
                "[",
                "]",
                "@",
                "!",
                "$",
                "'",
                "(",
                ")",
                "*",
                "+",
                ",",
                "=",
            ]:  # unreserved per RFC 3986 sec. 2.3
                res += c
            else:
                o = ord(c) & 0xFF
                res += f"%{o:02X}"

    return res


class Pkcs11URI:
    def __init__(self):
        self.reset()
        self.module_directories = []
        self.allowed_module_paths = []
        self.allow_any_module = True

    def reset(self) -> None:
        """Clear path and query parameters of the URI."""
        self.path_attributes = {}
        self.query_attributes = {}

    def get_attribute_bytes(
        self, attr_map: dict[str, str], name: str
    ) -> Optional[bytes]:
        """Get an attribute in form of bytes, if available."""
        if name not in attr_map:
            return None
        vb = b""
        for c in attr_map[name]:
            vb += ord(c).to_bytes(length=1, byteorder="big")
        return vb

    def get_path_attribute_bytes(self, name: str) -> Optional[bytes]:
        """Get a path attribute as <bytes>."""
        return self.get_attribute_bytes(self.path_attributes, name)

    def get_query_attribute_bytes(self, name: str) -> Optional[bytes]:
        """Get a query attribute as <bytes>."""
        return self.get_attribute_bytes(self.query_attributes, name)

    def set_attribute(
        self,
        attr_map: dict[str, str],
        name: str,
        value: str,
        is_path: bool = False,
        pctencode: bool = False,
    ) -> None:
        if pctencode:
            value = escape(value, is_path)
        """Set an attribute in the given dictionary."""
        # RFC7512 section 2.3: only 'id' may contain non-textual data
        # However, URI on page 15 shows a token with 'acute:%20>>%C3%A1<<' in
        # the name.
        # -> Only 'id' is relevant for user to have as bytes, all other ones can
        #    be given back as strings or pct-encoded
        vb = parse.unquote_to_bytes(value)

        v = ""
        for c in vb:
            v += chr(c)
        attr_map[name] = v

    def get_path_attribute(
        self, name: str, pctencode: bool = False
    ) -> Optional[str]:
        """Get a path attribute by its name, possibly in pct-encode form."""
        v = self.path_attributes.get(name)
        if v is not None and pctencode:
            v = escape(v, True)
        return v

    def set_path_attribute(
        self, name: str, value: str, pctencode: bool = False
    ) -> None:
        """Set a path attribute."""
        self.set_attribute(self.path_attributes, name, value, True, pctencode)

    def set_path_attribute_unencoded(self, name: str, value: bytes) -> None:
        """Set an unencoded path attribute as bytes."""
        self.set_path_attribute(name, value.decode("utf-8"), True)

    def add_path_attribute(
        self, name: str, value: str, pctencode: bool = False
    ) -> None:
        """Add a path attrbuute given as a string."""
        if name in self.path_attributes:
            raise ValueError("duplicate path attribute")
        self.set_path_attribute(name, value, pctencode)

    def add_path_attribute_unencoded(self, name: str, value: bytes) -> None:
        """Add an unencoded path attribute as bytes."""
        if name in self.path_attributes:
            raise ValueError("duplicate path attribute")
        self.set_path_attribute_unencoded(name, value)

    def remove_path_attribute(self, name: str) -> None:
        """Remove a path attribute."""
        if name in self.path_attributes:
            del self.path_attributes[name]

    def get_query_attribute(
        self, name: str, pctencode: bool = False
    ) -> Optional[str]:
        """Get a query attribute by its name, possibly in pct-encode form."""
        v = self.query_attributes.get(name)
        if v is not None and pctencode:
            v = escape(v, False)
        return v

    def set_query_attribute(
        self, name: str, value: str, pctencode: bool = False
    ) -> None:
        """Set a query attribute."""
        self.set_attribute(self.query_attributes, name, value, False, pctencode)

    def set_query_attribute_unencoded(self, name: str, value: bytes) -> None:
        """Set an unencoded query attribute as bytes."""
        self.set_query_attribute(name, value.decode("utf-8"), True)

    def add_query_attribute(
        self, name: str, value: str, pctencode: bool = False
    ) -> None:
        """Add a query attribute given as a string. pctencode it if wanted."""
        if name in self.query_attributes:
            raise ValueError("duplicate query attribute")
        self.set_query_attribute(name, value, pctencode)

    def add_query_attribute_unencoded(self, name: str, value: bytes) -> None:
        """Add an unencoded query attribute as bytes."""
        if name in self.query_attributes:
            raise ValueError("duplicate query attribute")
        self.set_query_attribute_unencoded(name, value)

    def remove_query_attribute(self, name: str) -> None:
        """Remove a query attribute."""
        if name in self.query_attributes:
            del self.query_attributes[name]

    def validate(self) -> None:
        """Validate a Pkcs11URI.

        Validate a Pkcs11URI object's attributes following RFC 7512 rules and
        proper formatting of their values.
        """
        v = self.path_attributes.get("slot-id")
        if v is not None and not any(i.isdigit() for i in v):
            raise ValueError(f"slot-id must be a number: {v}")

        v = self.path_attributes.get("library-version")
        if v is not None and re.match("^[0-9]+(\\.[0-9]+)?$", v) is None:
            raise ValueError(f"Invalid format for library-version '{v}'")

        v = self.path_attributes.get("type")
        if (
            v is not None
            and re.match("^(public|private|cert|secret-key|data)?$", v) is None
        ):
            raise ValueError(f"Invalid type '{v}'")

        v1 = self.query_attributes.get("pin-source")
        v2 = self.query_attributes.get("pin-value")
        if v1 is not None and v2 is not None:
            raise ValueError("URI must not contain pin-source and pin-value")

        v = self.query_attributes.get("module-path")
        if v is not None and not os.path.isabs(v):
            raise ValueError(
                f"path {v} of module-path attribute must be absolute"
            )

    def has_pin(self) -> bool:
        """Check whether a PIN has been provided."""
        return (
            self.query_attributes.get("pin-value") is not None
            or self.query_attributes.get("pin-source") is not None
        )

    def get_pin(self) -> str:
        """Get the PIN to access an object for example."""
        v = self.query_attributes.get("pin-value")
        if v is not None:
            return v

        v = self.query_attributes.get("pin-source")
        if v is not None:
            up = urlparse(v)
            if up.scheme in ["", "file"]:
                if not os.path.isabs(up.path):
                    raise ValueError(
                        f"PIN URI path '{up.path}' is not absolute"
                    )
                with open(up.path) as f:
                    return f.read()
            else:
                raise ValueError(
                    f"PIN URI scheme {up.scheme} is not supported: {{v}}"
                )

        raise ValueError("Neither pin-source nor pin-value are available")

    def parse(self, uri: str) -> None:
        """Parse the given URI."""
        if not uri.startswith("pkcs11:"):
            raise ValueError(
                f"Malformed pkcs11 URI: missing 'pkcs11:' prefix: {uri}"
            )

        self.reset()

        parts = uri[7:].split("?", 1)

        if len(parts[0]) > 0:
            # parse path part
            for part in parts[0].split(";"):
                p = part.split("=", 1)
                if len(p) != 2:
                    raise ValueError(
                        "Malformed pkcs11 URI: malformed path attribute"
                    )
                self.add_path_attribute(p[0], p[1])

        if len(parts) == 2:
            for part in parts[1].split("&"):
                p = part.split("=", 1)
                if len(p) != 2:
                    raise ValueError(
                        "Malformed pkcs11 URI: malformed query attribute"
                    )
                self.add_query_attribute(p[0], p[1])

        self.validate()

    def format_attributes(self, attr_map: dict[str, str], is_path: bool) -> str:
        """Format the attributes in the given map."""
        res = ""
        for key, value in attr_map.items():
            value = escape_all(value) if key == "id" else escape(value, is_path)
            if len(res) > 0:
                if is_path:
                    res += ";"
                else:
                    res += "&"
            res += key + "=" + value
        return res

    def format(self) -> str:
        """Format the Pkcs11URI to a string."""
        self.validate()
        result = "pkcs11:" + self.format_attributes(self.path_attributes, True)
        if len(self.query_attributes) > 0:
            result += "?" + self.format_attributes(self.query_attributes, False)
        return result

    def __str__(self) -> str:
        return self.format()

    def set_module_directories(self, dirs: Iterable[str]) -> None:
        """Set directories to search for pkcs11 modules."""
        self.module_directories = dirs

    def set_allowed_module_paths(self, allowed_paths: list[str]) -> None:
        """Set the allowed paths for pkcs11 modules."""
        self.allowed_module_paths = allowed_paths

    def set_allow_any_module(self, allow_any_module: bool) -> None:
        """Set the any module may be loaded."""
        self.allow_any_module = allow_any_module

    def _is_allowed_path(self, path: str, allowed_paths: list[str]) -> bool:
        """Check whether the given path is allowed."""
        if self.allow_any_module:
            return True
        for allowed_path in allowed_paths:
            if allowed_path == path:
                return True
            print(allowed_path)
            if allowed_path[-1] == os.path.sep and path.find(allowed_path) == 0:
                try:
                    path[len(allowed_path) :].index(os.path.sep)
                except ValueError:
                    return True
        return False

    def get_module(self) -> str:
        """Get the pkcs11 module that is to be used."""
        v = self.get_query_attribute("module-path")
        if v is not None:
            try:
                statbuf = stat(v)
                if S_ISREG(statbuf.st_mode):
                    if self._is_allowed_path(v, self.allowed_module_paths):
                        return v
                    raise ValueError(
                        f"module-path '{v}' is not allowed by policy"
                    )
                if not os.path.isdir(v):
                    raise ValueError(
                        f"module-path '{v}' points to an invlaid file type"
                    )
                searchdirs = [v]

            except FileNotFoundError as err:
                raise ValueError from err
        else:
            searchdirs = self.module_directories

        module_name = self.get_query_attribute("module-name")
        if module_name is None:
            raise ValueError("module-name attribute is not set")
        module_name = module_name.lower()

        for dir in searchdirs:
            files = glob.glob(os.path.join(dir, "*"))
            for file in files:
                file_lower = file.lower()
                try:
                    i = file_lower.index(module_name)
                except ValueError:
                    continue

                if (
                    len(file_lower) == i + len(module_name)
                    or file_lower[i + len(module_name)] == "."
                ):
                    f = os.path.join(dir, file)
                    if self._is_allowed_path(f, self.allowed_module_paths):
                        return f
                    raise ValueError(f"module '{f}' is not allowed by policy")
        dirs = ", ".join(searchdirs)
        raise ValueError(f"No module '{module_name}' could be found in {dirs}")

    def get_keyid_and_label(self) -> tuple[Optional[bytes], Optional[str]]:
        """Get the id for the key and its label."""
        keyid = self.get_path_attribute_bytes("id")
        label = self.get_path_attribute("object")
        if keyid is None and label is None:
            raise ValueError(
                "Neither 'id' nor 'object' attributes were found in pkcs11 URI"
            )
        return keyid, label

    def open_session(
        self,
        lib: PyKCS11.PyKCS11Lib,
        slot_id: int,
        pin: str,
        token_label: Optional[str],
    ) -> PyKCS11.Session:
        """Open a session.

        Open a session given a slot-id and an optional token label whose name
        must match if given.
        """
        token = lib.getTokenInfo(slot_id)
        if token_label is not None and token.label != f"{token_label:<32s}":
            raise ValueError(
                f"The token in slot {slot_id} is not called '{token_label}'"
            )
        session = lib.openSession(slot_id)
        if pin is not None:
            session.login(pin)
        return session

    def get_login_parameters(self) -> tuple[Optional[str], str, int]:
        """Get the login parameters PIN, module, and slot-id from the URI."""
        pin = None
        if self.has_pin():
            pin = self.get_pin()

        module = self.get_module()

        slotid = -1
        v = self.get_path_attribute("slot-id")
        if v is not None:
            slotid = int(v)
            if slotid < 0:
                raise ValueError("slot-id is a negative number")
            if slotid > 0xFFFFFFFF:
                raise ValueError("slot-id is larger than 32 bit")

        return pin, module, slotid

    def login(self) -> tuple[PyKCS11.Session, PyKCS11.PyKCS11Lib]:
        """Log in to the device using parameters from the URI."""
        pin, module, slot_id = self.get_login_parameters()

        lib = PyKCS11.PyKCS11Lib().load(pkcs11dll_filename=module)

        token_label = self.get_path_attribute("token")

        # If a slot-id is given, use it to open the session
        if slot_id >= 0:
            return self.open_session(lib, slot_id, pin, token_label), lib

        if token_label is None:
            raise ValueError("Need a token due to missing slot-id")

        for slot in lib.getSlotList():
            token = lib.getTokenInfo(slot)
            if token.label == f"{token_label:<32s}":
                session = lib.openSession(slot)
                if pin is not None:
                    session.login(pin)
                return session, lib
        raise ValueError(
            f"Could not find a token with label {token_label} in any slots"
        )
